import {
  Component,
  Input,
  OnInit,
  EventEmitter,
  Output,
  ViewChild,
  OnDestroy,
  ElementRef,
  AfterContentInit,
  ChangeDetectorRef,
  HostListener
} from '@angular/core';

import { Observable, of, Subscription, combineLatest, Subject } from 'rxjs';
import { catchError, map, tap, exhaustMap, take, filter } from 'rxjs/operators';
import { Store, select } from '@ngrx/store';

import * as d3 from 'd3';

import { Zone, Host, Service, Port } from '@overflow/model/discovery';

import * as AppStore from '../../../store';
import * as DiscoveryConfigStore from '../../../store/discovery/config';
import * as InfraStore from '../../../store/infra/infra';
import * as UILayoutStore from '../../../store/ui/layout';
import { DiscoverySession } from '../../../core/discovery/discovery-session';
import { DiscoveryMessageType } from 'src/app/core/type';
import { ConfirmationService } from 'primeng/primeng';
import { MetaServiceTypeEnum, toMetaServiceTypeEnum } from '@overflow/model/meta/MetaServiceType';

export class Link implements d3.SimulationLinkDatum<Node> {
  source: Node;
  target: Node;

  index?: number;

  constructor(source: Node, target: Node) {
    target.addLink(this);

    if ('zone' !== source.group) {
      source.addLink(this);
    }

    this.source = source;
    this.target = target;
  }
}

export class Node implements d3.SimulationNodeDatum {
  /**
   * Node’s zero-based index into nodes array. This property is set during the initialization process of a simulation.
   */
  index?: number;
  /**
   * Node’s current x-position
   */
  x?: number;
  /**
   * Node’s current y-position
   */
  y?: number;
  /**
   * Node’s current x-velocity
   */
  vx?: number;
  /**
   * Node’s current y-velocity
   */
  vy?: number;
  /**
   * Node’s fixed x-position (if position was fixed)
   */
  fx?: number | null;
  /**
   * Node’s fixed y-position (if position was fixed)
   */
  fy?: number | null;

  r?: number;

  id: string;

  group?: string;

  infra?: Host | Port | Service | null;

  image?: string;

  links?: Link[];

  constructor(id) {
    this.id = id;
    this.links = [];
  }

  addLink(link: Link) {
    this.links.push(link);
  }

}


export class DisplaySummary {
  totalHosts: number;
  totalPorts: number;
  totalServices: number;
  elapsedTime: string;

  startDate: Date;
  stopped: boolean;

  hTimer: any;

  constructor() {
    this.stopped = true;
    this.totalHosts = 0;
    this.totalPorts = 0;
    this.totalServices = 0;
    this.elapsedTime = '00:00:00';
  }

  start(startDate: Date) {
    this.stopped = false;
    this.startDate = startDate;
    const __this = this;
    const _startDate = new Date();
    this.hTimer = setInterval(function () {
      // Get todays date and time
      const now = new Date().getTime();
      __this.setElapsedTime(_startDate, new Date());
    }, 1000);
  }

  setElapsedTime(startDate: Date, now: Date) {
    const distance = now.getTime() - startDate.getTime();

    // Time calculations for days, hours, minutes and seconds
    const hours = Math.floor((distance % (1000 * 60 * 60 * 24)) / (1000 * 60 * 60));
    const minutes = Math.floor((distance % (1000 * 60 * 60)) / (1000 * 60));
    const seconds = Math.floor((distance % (1000 * 60)) / 1000);

    this.elapsedTime = String(hours).padStart(2, '0') + ':' + String(minutes).padStart(2, '0') + ':' + String(seconds).padStart(2, '0');
  }

  stop(stopDate: Date) {
    clearInterval(this.hTimer);
    this.setElapsedTime(this.startDate, stopDate);
    this.stopped = true;
  }

  increaseHost() {
    this.totalHosts++;
  }

  increasePort() {
    this.totalPorts++;
  }

  increaseService() {
    this.totalServices++;
  }
}

const FIRE_CLICK_DETAIL = -99999;

@Component({
  selector: 'app-infra-display-map',
  templateUrl: './map.component.html',
  styleUrls: ['./map.component.scss']
})
export class MapComponent implements OnInit, AfterContentInit, OnDestroy {

  zone: Zone | null = null;
  hosts: Map<string, Host>;
  ports: Map<string, Map<number, Port>>;
  services: Map<string, Map<number, Map<string, Service>>>;

  zoneNode: Node | null = null;
  nodes: Node[];
  links: Link[];
  nodeElements: Map<string, Element>;

  selectedNode: Node | null = null;

  @ViewChild('displayInfra') set displayInfra(content: ElementRef) {
    this.displayInfraRef = content;
    this.store.dispatch(new InfraStore.ChangeInfraDisplayElementRef({ infraDisplayElementRef: content }));
  }

  private displayInfraRef: ElementRef;
  public simulation: d3.Simulation<Node, Link> | undefined;
  private zoomBehavior: d3.ZoomBehavior<Element, {}>;
  private displayInfraWidth: number;
  private displayInfraHeight: number;
  private readonly maxScale: number;
  private readonly minScale: number;

  displaySummary: DisplaySummary | null = null;

  discoverySessionSubscription: Subscription | null = null;
  refreshInfraDisplaySubscription: Subscription | null = null;
  infraSubscription: Subscription | null = null;
  infraObservable: Observable<void>;
  selectedInfraSubscription: Subscription | null = null;

  discoverySubscription: Subscription | null = null;

  private preventBrowserCloseSubject: Subject<void>;
  private preventBrowserCloseSubscription: Subscription | null = null;

  constructor(
    private store: Store<any>,
    private changeDetector: ChangeDetectorRef,
    private confirmationService: ConfirmationService,
  ) {
    this.preventBrowserCloseSubject = new Subject<void>();

    this.maxScale = 1;
    this.minScale = 0.7;

    const __this = this;
    this.infraObservable = this.store.pipe(
      take(1),
      select((state) => {
        const zone = AppStore.DiscoverySelector.ConfigSelector.selectZone(state);
        const hosts = AppStore.InfraSelector.InfraSelector.selectHosts(state);
        const ports = AppStore.InfraSelector.InfraSelector.selectPorts(state);
        const services = AppStore.InfraSelector.InfraSelector.selectServices(state);

        __this.initMapDisplay(zone, hosts, ports, services);
      }),
    );
  }

  @HostListener('window:beforeunload', ['$event'])
  onBeforeUnload($event: Event) {
    if (null !== this.displaySummary && !this.displaySummary.stopped) {
      this.preventBrowserCloseSubject.next();
      $event.returnValue = true;
    }
  }

  @HostListener('window:resize', ['$event'])
  onResize(event) {
    if (undefined !== this.displayInfraRef) {
      this.displayInfraWidth = this.displayInfraRef.nativeElement.clientWidth;
      this.displayInfraHeight = this.displayInfraRef.nativeElement.clientHeight;

      this.zoneNode.fx = this.displayInfraWidth / 2;
      this.zoneNode.fy = this.displayInfraHeight / 2;

      this.simulationRestart(false);
    }
  }


  ngOnInit(): void {
    const __this = this;
    this.discoverySessionSubscription = this.store.pipe(
      select(AppStore.DiscoverySelector.RequestSelector.selectDiscoverySession),
      map((discoverySession: DiscoverySession) => {
        if (null === discoverySession) {
          if (null !== __this.discoverySubscription) {
            __this.discoverySubscription.unsubscribe();
            __this.discoverySubscription = null;
          }
          return;
        }
        __this.discoverySubscription = discoverySession.discoveryObservable().pipe(
          map((discoveryMessage: { messageType: DiscoveryMessageType, params: any[] }) => {
            switch (discoveryMessage.messageType) {
              case DiscoveryMessageType.Queueing:
                break;
              case DiscoveryMessageType.QueueingFailed:
                break;
              case DiscoveryMessageType.QueueingTimeout:
                break;
              case DiscoveryMessageType.DiscoveryStart: {
                __this.store.dispatch(new UILayoutStore.ChangeBlockPagesContent({ blockPagesContent: true }));
                __this.initMapDisplay(__this.zone, null, null, null);
                __this.displaySummary.start(discoveryMessage.params[0]);
              }
                break;
              case DiscoveryMessageType.DiscoveryStop: {
                __this.zone.hostList = [];
                __this.hosts.forEach(_host => {
                  if (_host.zone.network === __this.zone.network) {
                    __this.zone.hostList.push(_host);
                  }
                  _host.portList = [];
                  if (__this.ports.has(_host.address)) {
                    __this.ports.get(_host.address).forEach(_port => {
                      _host.portList.push(_port);

                      _port.serviceList = [];
                      if (__this.services.has(_host.address) && __this.services.get(_host.address).has(_port.portNumber)) {
                        __this.services.get(_host.address).get(_port.portNumber).forEach(_service => {
                          _port.serviceList.push(_service);
                        });
                      }
                    });
                  }
                });

                __this.simulationRestart(true);
                __this.zoomToFit(0.95, 500);

                __this.displaySummary.stop(discoveryMessage.params[0]);
                __this.store.dispatch(new UILayoutStore.ChangeBlockPagesContent({ blockPagesContent: false }));
              }
                break;
              case DiscoveryMessageType.DiscoveryMode:
                break;
              case DiscoveryMessageType.DiscoveryError:
                break;
              case DiscoveryMessageType.DiscoveredHost:

                __this.addHost(discoveryMessage.params[0], true);
                break;
              case DiscoveryMessageType.DiscoveredPort:
                __this.addPort(discoveryMessage.params[0], false);
                break;
              case DiscoveryMessageType.DiscoveredService:
                __this.addService(discoveryMessage.params[0], true);
                break;
              default:
                break;
            }
          }),
        ).subscribe();
      }),
      catchError(error => {
        console.log(error);
        return of();
      }),
    ).subscribe();

    this.refreshInfraDisplaySubscription = this.store.pipe(
      select(AppStore.InfraSelector.InfraSelector.selectRefreshInfraDisplay),
      filter(v => !!v),
      map((refreshInfraDisplay: boolean) => {
        if (!refreshInfraDisplay) {
          return;
        }
        __this.infraObservable.subscribe();
      }),
      catchError(error => {
        console.log(error);
        return of();
      }),
    ).subscribe();

    this.selectedInfraSubscription = this.store.pipe(
      select(AppStore.InfraSelector.InfraSelector.selectSelectedInfra)
    ).pipe(
      map((selectedInfra) => {
        if (null === selectedInfra) {
          __this.onHideDetail();
          return;
        } else {
          __this.onInfraSelected(selectedInfra);
        }
      }),
    ).subscribe();

    this.preventBrowserCloseSubscription = this.preventBrowserCloseSubject
      .pipe(
        map(() => {
          this.confirmationService.confirm({
            message: 'Discovery is in progress.',
            header: 'Alert',
            icon: 'pi pi-info-circle',
            acceptLabel: 'Yes',
            acceptVisible: true,
            rejectVisible: false,
          });
        }),
        catchError(error => {
          console.log(error);
          alert('An error has occurred.');
          return of();
        }),
      )
      .subscribe();

  }

  ngAfterContentInit(): void {
    this.displayInfraWidth = this.displayInfraRef.nativeElement.clientWidth;
    this.displayInfraHeight = this.displayInfraRef.nativeElement.clientHeight;

    this.infraObservable.subscribe();
  }

  ngOnDestroy(): void {
    this.store.dispatch(new InfraStore.ChangeInfraDisplayElementRef({ infraDisplayElementRef: null }));

    if (null !== this.infraSubscription) {
      this.infraSubscription.unsubscribe();
    }
    if (null !== this.refreshInfraDisplaySubscription) {
      this.refreshInfraDisplaySubscription.unsubscribe();
    }
    if (null !== this.discoverySessionSubscription) {
      this.discoverySessionSubscription.unsubscribe();
    }
    if (null !== this.selectedInfraSubscription) {
      this.selectedInfraSubscription.unsubscribe();
    }
    if (null !== this.preventBrowserCloseSubscription) {
      this.preventBrowserCloseSubscription.unsubscribe();
    }
  }

  initMapDisplay(zone: Zone, hosts: Host[], ports: Port[], services: Service[]) {
    if (null === zone) {
      return;
    }
    this.nodes = [];
    this.links = [];
    this.nodeElements = new Map();
    this.hosts = new Map();
    this.ports = new Map();
    this.services = new Map();
    this.displaySummary = new DisplaySummary();

    this.simulationInit();

    this.setZone(zone);

    if (null !== hosts) {
      hosts.forEach((host) => {
        this.addHost(host);
      });
    }

    if (null !== ports) {
      ports.forEach((port) => {
        this.addPort(port);
      });
    }

    if (null !== services) {
      services.forEach((service) => {
        this.addService(service);
      });
    }

    this.simulationRestart(true);
  }

  simulationInit() {
    if (undefined !== this.simulation) {
      return;
    }

    const svg = d3.select(this.displayInfraRef.nativeElement);

    this.zoomBehavior = d3.zoom()
      .scaleExtent([0.2, 4])
      .on('zoom', () => {
        const transform = d3.event.transform;
        svg.select('g').attr('transform', 'translate(' + transform.x + ',' + transform.y + ') scale(' + transform.k + ')');
      });
    svg.call(this.zoomBehavior);

    const __this = this;
    this.simulation = d3
      .forceSimulation<Node, Link>()
      .nodes(this.nodes)
      .force('charge', d3.forceManyBody().strength(-200))
      .force('link', d3.forceLink(this.links).distance(150))
      .force('center', d3.forceCenter(this.displayInfraWidth / 2, this.displayInfraHeight / 2))
      .force('collision', d3.forceCollide().radius(function (node: Node) {
        return node.r * 1.2;
      }))
      .on('tick', () => {
        __this.changeDetector.markForCheck();
      })
      ;
  }

  simulationRestart(attachEvent: boolean = false) {
    // Update and restart the simulation.
    this.simulation
      .nodes(this.nodes)
      .force('link', d3.forceLink(this.links).distance(150))
      .force('center', d3.forceCenter(this.displayInfraWidth / 2, this.displayInfraHeight / 2))
      .alpha(1)
      .restart()
      ;
    this.changeDetector.detectChanges();

    if (attachEvent) {
      const __this = this;

      const nodeElements = d3.select(this.displayInfraRef.nativeElement).selectAll('.node-container');
      const linkElements = d3.select(this.displayInfraRef.nativeElement).selectAll('.link-container');

      function getNodeFromElement(element: Element): Node | null {
        const container = d3.select(element);
        const nodeId = container.attr('nodeId');
        const { node } = __this.getNode(nodeId);
        return node;
      }

      // drag
      nodeElements.call(
        d3.drag()
          .on('start', function () {
            const node = getNodeFromElement(this);
            if (null === node || 'zone' === node.group) {
              return;
            }

            d3.event.sourceEvent.stopPropagation();

            if (!d3.event.active) {
              __this.simulation.alphaTarget(0.3).restart();
            }

            d3.event.on('drag', dragged).on('end', ended);

            function dragged(d, i) {
              node.fx = d3.event.x;
              node.fy = d3.event.y;
            }

            function ended() {
              if (!d3.event.active) {
                __this.simulation.alphaTarget(0);
              }

              node.fx = null;
              node.fy = null;
            }
          })
      );

      enum ConnectedNode {
        TARGET = 1,
        CONNECTED,
        NOT_CONNECTED,
      }

      function isConnectedNode(node: Node, element: Element): ConnectedNode {
        const _node = getNodeFromElement(element);
        if (node === _node) {
          return ConnectedNode.TARGET;
        }
        for (let i = 0; i < node.links.length; i++) {
          const link = node.links[i];
          if (node === _node || link.source === _node || link.target === _node) {
            return ConnectedNode.CONNECTED;
          }
        }
        return ConnectedNode.NOT_CONNECTED;
      }

      function isConnectedLink(node: Node, element: Element) {
        const _linkElement = d3.select(element);
        const sourceId = _linkElement.attr('sourceId');
        const targetId = _linkElement.attr('targetId');

        if (sourceId === node.id || targetId === node.id) {
          return true;
        }

        return false;
      }

      nodeElements.each(function (d) {
        const nodeElement = this as Element;
        const node = getNodeFromElement(nodeElement);
        __this.nodeElements.set(node.id, nodeElement);
      });

      nodeElements.on('click', function () {
        d3.event.stopPropagation();

        const nodeElement = this as Element;
        const node = getNodeFromElement(nodeElement);
        if (null === node || 'zone' === node.group) {
          clearSelection();
          if (FIRE_CLICK_DETAIL !== d3.event.detail) {
            __this.onInfraClick(node);
          }
          return;
        }

        if (null === __this.selectedNode) {
          nodeElements.each(function () {
            const _thisElement = this as Element;
            d3.select(_thisElement).classed('semi-unfocused', false);
            d3.select(_thisElement).classed('unfocused', false);
          });

          linkElements.each(function () {
            const _thisElement = this as Element;
            d3.select(_thisElement).classed('semi-unfocused', false);
            d3.select(_thisElement).classed('unfocused', false);
          });
        } else {
          nodeElements.each(function () {
            const _thisElement = this as Element;
            d3.select(_thisElement).classed('semi-unselected', false);
            d3.select(_thisElement).classed('unselected', false);
          });

          linkElements.each(function () {
            const _thisElement = this as Element;
            d3.select(_thisElement).classed('semi-unselected', false);
            d3.select(_thisElement).classed('unselected', false);
          });
        }

        __this.selectedNode = node;

        nodeElements.each(function () {
          const _thisElement = this as Element;
          switch (isConnectedNode(node, _thisElement)) {
            case ConnectedNode.CONNECTED:
              d3.select(_thisElement).classed('semi-unselected', true);
              break;
            case ConnectedNode.NOT_CONNECTED:
              d3.select(_thisElement).classed('unselected', true);
              break;
            default:
              break;
          }
        });

        linkElements.each(function () {
          const _thisElement = this as Element;
          if (isConnectedLink(node, _thisElement)) {
            d3.select(_thisElement).classed('semi-unselected', true);
          } else {
            d3.select(_thisElement).classed('unselected', true);
          }
        });

        if (FIRE_CLICK_DETAIL !== d3.event.detail) {
          __this.onInfraClick(node);
        }
      });

      // Highlight
      function clearSelection() {
        nodeElements.each(function () {
          const _thisElement = this as Element;
          d3.select(_thisElement).classed('semi-unselected', false);
          d3.select(_thisElement).classed('unselected', false);
        });

        linkElements.each(function () {
          const _thisElement = this as Element;
          d3.select(_thisElement).classed('semi-unselected', false);
          d3.select(_thisElement).classed('unselected', false);
        });

      }
      const displayInfra = d3.select(this.displayInfraRef.nativeElement);
      displayInfra.on('click', function () {
        __this.selectedNode = null;
        __this.store.dispatch(new InfraStore.ChangeSelectedInfra(null));

        clearSelection();
      });

      nodeElements
        .on('mouseover', function () {
          const node = getNodeFromElement(this as Element);
          if ('zone' === node.group) {
            return;
          }

          if (null === __this.selectedNode) {
            nodeElements.each(function () {
              const _thisElement = this as Element;
              switch (isConnectedNode(node, _thisElement)) {
                case ConnectedNode.CONNECTED:
                  d3.select(_thisElement).classed('semi-unfocused', true);
                  break;
                case ConnectedNode.NOT_CONNECTED:
                  d3.select(_thisElement).classed('unfocused', true);
                  break;
                default:
                  break;
              }
            });

            linkElements.each(function () {
              const _thisElement = this as Element;
              if (isConnectedLink(node, _thisElement)) {
                d3.select(_thisElement).classed('semi-unfocused', true);
              } else {
                d3.select(_thisElement).classed('unfocused', true);
              }
            });
          } else {
            nodeElements.each(function () {
              const _thisElement = this as Element;
              switch (isConnectedNode(node, _thisElement)) {
                case ConnectedNode.TARGET:
                  d3.select(_thisElement).classed('focused', true);
                  break;
                case ConnectedNode.CONNECTED:
                  d3.select(_thisElement).classed('semi-focused', true);
                  break;
                default:
                  break;
              }
            });

            linkElements.each(function () {
              const _thisElement = this as Element;
              if (isConnectedLink(node, _thisElement)) {
                d3.select(_thisElement).classed('semi-focused', true);
              }
            });
          }

        })
        .on('mouseout', function () {
          const node = getNodeFromElement(this as Element);

          if (null === __this.selectedNode) {
            nodeElements.each(function () {
              const _thisElement = this as Element;
              d3.select(_thisElement).classed('semi-unfocused', false);
              d3.select(_thisElement).classed('unfocused', false);
            });

            linkElements.each(function () {
              const _thisElement = this as Element;
              d3.select(_thisElement).classed('semi-unfocused', false);
              d3.select(_thisElement).classed('unfocused', false);
            });

          } else {
            nodeElements.each(function () {
              const _thisElement = this as Element;
              d3.select(_thisElement).classed('semi-focused', false);
              d3.select(_thisElement).classed('focused', false);
            });

            linkElements.each(function () {
              const _thisElement = this as Element;
              d3.select(_thisElement).classed('semi-focused', false);
              d3.select(_thisElement).classed('focused', false);
            });

          }
        })
        ;

    }
  }

  zoomToFit(paddingPercent, transitionDuration) {
    const root = d3.select(this.displayInfraRef.nativeElement);
    const bounds = root.node().getBBox();

    const fullWidth = this.displayInfraWidth;
    const fullHeight = this.displayInfraHeight;

    const width = bounds.width;
    const height = bounds.height;

    const midX = bounds.x + width / 2;
    const midY = bounds.y + height / 2;

    if (width === 0 || height === 0) { return; } // nothing to fit

    let scale = (paddingPercent || 0.75) / Math.max(width / fullWidth, height / fullHeight);
    console.log(`scale: ${scale}`);

    if (this.maxScale < scale) {
      scale = this.maxScale;
    }
    if (this.minScale > scale) {
      scale = this.minScale;
    }

    const translate = [fullWidth / 2 - scale * midX, fullHeight / 2 - scale * midY];

    root
      .transition()
      .duration(transitionDuration || 0) // milliseconds
      .call(this.zoomBehavior.transform, d3.zoomIdentity.translate(translate[0], translate[1]).scale(scale));
  }

  onInfraSelected({ group, infra }: { group: string, infra: Host | Port | Service }) {
    switch (group) {
      case 'zone': {
        const id = this.getZoneId(infra);
        this.fireClickOnNodeElement(id);
      }
        break;
      case 'host': {
        const id = this.getHostId(infra);
        this.fireClickOnNodeElement(id);
      }
        break;
      case 'service': {
        const id = this.getServiceId(infra);
        this.fireClickOnNodeElement(id);
      }

        break;
      default:
        break;
    }
  }


  fireClickOnNodeElement(nodeId: string) {
    const nodeElement = this.nodeElements.get(nodeId);
    if (null === nodeElement || (null !== this.selectedNode && nodeId === this.selectedNode.id)) {
      return;
    }
    nodeElement.dispatchEvent(new CustomEvent('click', { detail: FIRE_CLICK_DETAIL }));
  }

  onInfraClick(node: Node) {
    // switch (node.group) {
    //   case 'zone':
    //     const zone: Zone = node.target;
    //     zone.hostList = [];

    //     this.hosts.forEach(_host => {
    //       if (_host.zone.network === zone.network) {
    //         zone.hostList.push(_host);
    //       }
    //     });

    //     break;
    //   case 'host':
    //     const host: Host = node.target;
    //     host.portList = [];
    //     if (this.ports.has(host.address)) {
    //       this.ports.get(host.address).forEach(port => {
    //         host.portList.push(port);
    //       });
    //     }
    //     break;
    //   default:
    //     break;
    // }

    this.store.dispatch(new InfraStore.ChangeSelectedInfra({
      group: node.group,
      infra: node.infra,
    }));

    this.selectedNode = node;
  }

  onHideDetail() {
    const __this = this;

    const nodeElements = d3.select(this.displayInfraRef.nativeElement).selectAll('.node-container');
    const linkElements = d3.select(this.displayInfraRef.nativeElement).selectAll('.link-container');

    __this.selectedNode = null;

    nodeElements.each(function () {
      const _thisElement = this as Element;
      d3.select(_thisElement).classed('semi-unselected', false);
      d3.select(_thisElement).classed('unselected', false);
    });

    linkElements.each(function () {
      const _thisElement = this as Element;
      d3.select(_thisElement).classed('semi-unselected', false);
      d3.select(_thisElement).classed('unselected', false);
    });
  }


  private getNode(id: string): { index: number, node: Node } {
    let _n: Node = null;
    let _idx = -1;
    this.nodes.some((node): boolean => {
      _idx++;
      if (node.id === id) {
        _n = node;
        return true;
      }
      return false;
    });
    return { index: null === _n ? -1 : _idx, node: _n };
  }

  private getLink(source: Node, target: Node): { index: number, link: Link } | null {
    let _l: Link = null;
    let _idx = -1;
    this.links.some((link): boolean => {
      _idx++;
      if (link.source.id === source.id && link.target.id === target.id) {
        _l = link;
        return true;
      }
      return false;
    });
    return { index: null === _l ? -1 : _idx, link: _l };
  }

  refreshInfraDisplay(zone: Zone, hosts: Host[], services: Service[]) {

  }

  getZoneId(zone: Zone): string {
    return `${zone.network}`;
  }
  getHostId(host: Host): string {
    return `${this.getZoneId(host.zone)}-${host.address}`;
  }
  getPortId(port: Port): string {
    return `${this.getHostId(port.host)}-${port.portNumber}-${port.metaPortType.key}`;
  }
  getServiceId(service: Service): string {
    return `${this.getPortId(service.port)}-${service.serviceType}`;
  }
  getUnknownServiceId(service: Service): string {
    return `${this.getPortId(service.port)}-${MetaServiceTypeEnum.UNKNOWN}`;
  }

  setZone(zone: Zone, requireRefresh: boolean = false) {
    if (null === zone) {
      return;
    }
    this.zone = zone;
    this.zoneNode = new Node(this.getZoneId(zone));
    this.zoneNode.group = 'zone';
    this.zoneNode.infra = zone;
    this.zoneNode.fx = this.displayInfraWidth / 2;
    this.zoneNode.fy = this.displayInfraHeight / 2;
    this.zoneNode.r = 60;

    this.nodes.push(
      this.zoneNode,
    );

    if (requireRefresh) {
      this.simulationRestart();
    }
  }

  addHost(host: Host, requireRefresh: boolean = false) {
    if (null === host) {
      return;
    }
    this.hosts.set(host.address, host);

    const hostId = this.getHostId(host);

    let { node } = this.getNode(hostId);
    if (null !== node) {
      node.infra = host;
    } else {
      node = new Node(hostId);
      node.infra = host;
      node.group = 'host';
      node.r = 40;
      node.x = this.zoneNode.x;
      node.y = this.zoneNode.y;

      this.nodes.push(node);
      this.links.push(new Link(this.zoneNode, node));

      this.displaySummary.increaseHost();

      if (requireRefresh) {
        this.simulationRestart();
      }
    }

  }

  addPort(port: Port, requireRefresh: boolean = false) {
    if (null === port) {
      return;
    }

    let _ports: Map<number, Port>;
    if (!this.ports.has(port.host.address)) {
      _ports = new Map();
      this.ports.set(port.host.address, _ports);
    } else {
      _ports = this.ports.get(port.host.address);
    }

    if (!_ports.has(port.portNumber)) {
      this.displaySummary.increasePort();
    }

    _ports.set(port.portNumber, port);

    if (requireRefresh) {
      this.simulationRestart();
    }
  }

  addService(service: Service, requireRefresh: boolean = false) {
    if (null === service) {
      return;
    }

    const hostId = this.getHostId(service.port.host);
    const { node: hostNode } = this.getNode(hostId);

    const unknownServiceId = this.getUnknownServiceId(service);
    const { index: unknownServiceNodeIdx, node: unknownServiceNode } = this.getNode(unknownServiceId);

    const serviceId = this.getServiceId(service);
    let { node: serviceNode } = this.getNode(serviceId);

    let newNode = false;

    if (MetaServiceTypeEnum.UNKNOWN === MetaServiceTypeEnum[service.serviceType]) {
      if (null !== unknownServiceNode) {
        unknownServiceNode.infra = service;
      } else {
        newNode = true;
      }
    } else {
      if (null !== serviceNode) {
        serviceNode.infra = service;
      } else {
        if (null !== unknownServiceNode) {
          const { index: linkIdx } = this.getLink(hostNode, unknownServiceNode);
          delete this.links[linkIdx];
          delete this.nodes[unknownServiceNodeIdx];
        }
        newNode = true;
      }
    }

    let _servicesMap: Map<number, Map<string, Service>>;
    if (!this.services.has(service.port.host.address)) {
      _servicesMap = new Map();
      this.services.set(service.port.host.address, _servicesMap);
    } else {
      _servicesMap = this.services.get(service.port.host.address);
    }

    let _services: Map<string, Service>;
    if (!_servicesMap.has(service.port.portNumber)) {
      _services = new Map();
      _servicesMap.set(service.port.portNumber, _services);
    } else {
      _services = _servicesMap.get(service.port.portNumber);
    }

    _services.set(service.serviceType, service);

    if (newNode) {
      serviceNode = new Node(serviceId);
      serviceNode.infra = service;
      serviceNode.group = 'service';
      serviceNode.r = 30;
      serviceNode.x = hostNode.x;
      serviceNode.y = hostNode.y;

      this.nodes.push(serviceNode);
      this.links.push(new Link(hostNode, serviceNode));

      this.displaySummary.increaseService();

      if (requireRefresh) {
        this.simulationRestart();
      }
    }
  }
}
